/*
 * Copyright © 2022 Soren Stoutner <soren@stoutner.com>.
 *
 * This file is part of Privacy Browser PC <https://www.stoutner.com/privacy-browser-pc>.
 *
 * Privacy Browser PC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Privacy Browser PC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Privacy Browser PC.  If not, see <http://www.gnu.org/licenses/>.
 */

// Application headers.
#include "BrowserView.h"
#include "MouseEventFilter.h"
#include "Settings.h"
#include "ui_BrowserView.h"
#include "UrlRequestInterceptor.h"
#include "dialogs/DomainSettingsDialog.h"
#include "helpers/DomainsDatabaseHelper.h"
#include "helpers/SearchEngineHelper.h"
#include "helpers/UserAgentHelper.h"
#include "windows/BrowserWindow.h"

// Qt framework headers.
#include <QAction>
#include <QWebEngineProfile>

BrowserView::BrowserView(QWidget *parent) : QWidget(parent)
{
    // Instantiate the browser view UI.
    Ui::BrowserView browserViewUi;

    // Setup the UI.
    browserViewUi.setupUi(this);

    // Get handles for the views.
    backButtonPointer = browserViewUi.backButton;
    forwardButtonPointer = browserViewUi.forwardButton;
    QPushButton *refreshButtonPointer = browserViewUi.refreshButton;
    QPushButton *homeButtonPointer = browserViewUi.homeButton;
    urlLineEditPointer = browserViewUi.urlLineEdit;
    javaScriptButtonPointer = browserViewUi.javaScript;
    QPushButton *domainSettingsButtonPointer = browserViewUi.domainSettingsButton;
    webEngineViewPointer = browserViewUi.webEngineView;

    // Get handles for the aspects of the WebEngine.
    QWebEnginePage *webEnginePagePointer = webEngineViewPointer->page();
    webEngineHistoryPointer = webEnginePagePointer->history();
    webEngineProfilePointer = webEnginePagePointer->profile();
    webEngineSettingsPointer = webEngineViewPointer->settings();

    // Update the webengine view from the URL line edit.  TODO.  Remove.
    connect(urlLineEditPointer, SIGNAL(returnKeyPressed(const QString)), this, SLOT(loadUrlFromLineEdit(const QString)));

    // Update the URL line edit from the webengine view.
    connect(webEngineViewPointer, SIGNAL(loadStarted()), this, SLOT(updateInterface()));
    connect(webEngineViewPointer, SIGNAL(loadProgress(const int)), this, SLOT(updateInterface()));
    connect(webEngineViewPointer, SIGNAL(loadFinished(const bool)), this, SLOT(updateInterface()));

    // Setup the URL bar buttons.  TODO.  Remove.
    connect(backButtonPointer, SIGNAL(clicked()), webEngineViewPointer, SLOT(back()));
    connect(forwardButtonPointer, SIGNAL(clicked()), webEngineViewPointer, SLOT(forward()));
    connect(refreshButtonPointer, SIGNAL(clicked()), webEngineViewPointer, SLOT(reload()));
    connect(homeButtonPointer, SIGNAL(clicked()), this, SLOT(home()));
    connect(javaScriptButtonPointer, SIGNAL(clicked()), this, SLOT(toggleJavaScript()));
    connect(domainSettingsButtonPointer, SIGNAL(clicked()), this, SLOT(openDomainSettings()));

    // Get the URL line edit palettes.  TODO.  Remove.
    noDomainSettingsPalette = urlLineEditPointer->palette();
    domainSettingsPalette = urlLineEditPointer->palette();

    // Modify the domain settings palette.  TODO.  Remove.
    domainSettingsPalette.setColor(QPalette::Base, QColor("#C8E6C9"));

    // Instantiate the mouse event pointer.
    MouseEventFilter *mouseEventFilterPointer = new MouseEventFilter(webEngineViewPointer);

    // Install the mouse event filter.
    qApp->installEventFilter(mouseEventFilterPointer);

    // Listen for hovered link URLs.
    connect(webEnginePagePointer, SIGNAL(linkHovered(const QString)), this, SLOT(pageLinkHovered(const QString)));

    // Instantiate the URL request interceptor.
    UrlRequestInterceptor *urlRequestInterceptorPointer = new UrlRequestInterceptor();

    // Set the URL request interceptor.
    webEngineProfilePointer->setUrlRequestInterceptor(urlRequestInterceptorPointer);

    // Reapply the domain settings when the host changes.
    connect(urlRequestInterceptorPointer, SIGNAL(applyDomainSettings(QString)), this, SLOT(applyDomainSettingsWithoutReloading(QString)));

    // Disable the cache.
    webEngineProfilePointer->setHttpCacheType(QWebEngineProfile::NoCache);

    // Don't allow JavaScript to open windows.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptCanOpenWindows, false);

    // Set the focus on the WebEngine view.
    webEngineViewPointer->setFocus();
}

void BrowserView::applyApplicationSettings()
{
    // Set the search engine URL.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(Settings::searchEngine());

    // Emit the search engine updated signal, which causes the on-the-fly menu to be updated.
    emit searchEngineUpdated(Settings::searchEngine());
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
void BrowserView::applyDomainSettingsAndReload() const
{
    // Apply the domain settings.  `true` reloads the website.
    applyDomainSettings(webEngineViewPointer->url().host(), true);
}

// This exists as a separate function from `applyDomainSettings()` so it can be listed as a slot and function without the need for a boolean argument.
void BrowserView::applyDomainSettingsWithoutReloading(const QString &hostname) const
{
    // Apply the domain settings  `false` does not reload the website.
    applyDomainSettings(hostname, false);
}

void BrowserView::applyDomainSettings(const QString &hostname, const bool reloadWebsite) const
{
    // Get the record for the hostname.
    QSqlQuery domainQuery = DomainsDatabaseHelper::getDomainQuery(hostname);

    // Check if the hostname has domain settings.
    if (domainQuery.isValid())  // The hostname has domain settings.
    {
        // Get the domain record.
        QSqlRecord domainRecord = domainQuery.record();

        // Set the JavaScript status.
        switch (domainRecord.field(DomainsDatabaseHelper::JAVASCRIPT).value().toInt())
        {
            case (DomainsDatabaseHelper::SYSTEM_DEFAULT):
            {
                // Set the default JavaScript status.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScript());

                break;
            }

            case (DomainsDatabaseHelper::DISABLED):
            {
                // Disable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, false);

                break;
            }

            case (DomainsDatabaseHelper::ENABLED):
            {
                // Enable JavaScript.
                webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, true);

                break;
            }
        }

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getResultingDomainSettingsUserAgent(domainRecord.field(DomainsDatabaseHelper::USER_AGENT).value().toString()));

        // Set the zoom factor.
        webEngineViewPointer->setZoomFactor(Settings::zoomFactor());

        // Apply the domain settings palette to the URL line edit.
        urlLineEditPointer->setPalette(domainSettingsPalette);  // TODO.  Remove.
        emit updateDomainSettingsIndicator(true);
    }
    else  // The hostname does not have domain settings.
    {
        // Set the JavaScript status.
        webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, Settings::javaScript());

        // Set the user agent.
        webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromDatabaseName(Settings::userAgent()));

        // Set the zoom factor.
        webEngineViewPointer->setZoomFactor(Settings::zoomFactor());

        // Apply the no domain settings palette to the URL line edit.
        urlLineEditPointer->setPalette(noDomainSettingsPalette);  // TODO.  Remove.
        emit updateDomainSettingsIndicator(false);
    }

    // Update the JavaScript button.
    if (webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled))
    {
        javaScriptButtonPointer->setIcon(QIcon(":/icons/javascript-warning"));
    }
    else
    {
        javaScriptButtonPointer->setIcon(QIcon(":/icons/privacy-mode"));
    }

    // Emit the on-the-fly menu update signals.
    emit userAgentUpdated(webEngineProfilePointer->httpUserAgent());
    emit zoomFactorUpdated(Settings::zoomFactor());

    // Reload the website if requested.
    if (reloadWebsite)
    {
        webEngineViewPointer->reload();
    }
}

void BrowserView::applyOnTheFlySearchEngine(QAction *searchEngineActionPointer)
{
    // Store the search engine name.
    QString searchEngineName = searchEngineActionPointer->text();

    // Strip out any `&` characters.
    searchEngineName.remove('&');

    // Store the search engine string.
    searchEngineUrl = SearchEngineHelper::getSearchUrl(searchEngineName);
}

void BrowserView::applyOnTheFlyUserAgent(QAction *userAgentActionPointer) const
{
    // Get the user agent name.
    QString userAgentName = userAgentActionPointer->text();

    // Strip out any `&` characters.
    userAgentName.remove('&');

    // Apply the user agent.
    webEngineProfilePointer->setHttpUserAgent(UserAgentHelper::getUserAgentFromTranslatedName(userAgentName));

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::applyOnTheFlyZoomFactor(const double &zoomFactor) const
{
    // Set the zoom factor.
    webEngineViewPointer->setZoomFactor(zoomFactor);
}

void BrowserView::back() const
{
    // Go back.
    webEngineViewPointer->back();
}

void BrowserView::forward() const
{
    // Go forward.
    webEngineViewPointer->forward();
}

void BrowserView::home() const
{
    // Load the homepage.
    webEngineViewPointer->load(QUrl::fromUserInput(Settings::homepage()));
}

void BrowserView::loadInitialWebsite()
{
    // Apply the application settings.
    applyApplicationSettings();

    // Get the arguments.
    QStringList argumentsStringList = qApp->arguments();

    // Check to see if the arguments lists contains a URL.
    if (argumentsStringList.size() > 1)
    {
        // Load the URL from the arguments list.
        webEngineViewPointer->load(QUrl::fromUserInput(argumentsStringList.at(1)));
    }
    else
    {
        // Load the homepage.
        home();
    }
}

void BrowserView::loadUrlFromLineEdit(QString url) const
{
    // Remove the focus from the URL line edit.  TODO.  Remove.
    urlLineEditPointer->clearFocus();

    // Decide if the text is more likely to be a URL or a search.
    if (url.contains("."))  // The text is likely a URL.
    {
        // Check if the URL does not start with a valid protocol.
        if (!url.startsWith("http") && !url.startsWith("file://"))
        {
            // Add `https://` to the beginning of the URL.
            url = "https://" + url;
        }

        // Load the URL.
        webEngineViewPointer->load(QUrl::fromUserInput(url));
    }
    else  // The text is likely a search.
    {
        // Load the search.
        webEngineViewPointer->load(QUrl::fromUserInput(searchEngineUrl + url));
    }
}

void BrowserView::openDomainSettings() const
{
    // Instantiate the domain settings window.
    DomainSettingsDialog *domainSettingsDialogPointer = new DomainSettingsDialog();

    // Set the dialog window title.
    domainSettingsDialogPointer->setWindowTitle(i18nc("The domain settings dialog title", "Domain Settings"));

    // Set the modality.
    domainSettingsDialogPointer->setWindowModality(Qt::WindowModality::WindowModal);;

    // Show the dialog.
    domainSettingsDialogPointer->show();

    // Reload the tabs when domain settings are updated.
    connect(domainSettingsDialogPointer, SIGNAL(domainSettingsUpdated()), this, SLOT(applyDomainSettingsAndReload()));
}

void BrowserView::pageLinkHovered(const QString &linkUrl) const
{
    // Emit a signal so that the browser window can update the status bar.
    emit linkHovered(linkUrl);
}

void BrowserView::refresh() const
{
    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::toggleJavaScript() const
{
    // Toggle JavaScript.
    webEngineSettingsPointer->setAttribute(QWebEngineSettings::JavascriptEnabled, !webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled));

    // Update the JavaScript button.
    if (webEngineSettingsPointer->testAttribute(QWebEngineSettings::JavascriptEnabled))
    {
        javaScriptButtonPointer->setIcon(QIcon(":/icons/javascript-warning"));
    }
    else
    {
        javaScriptButtonPointer->setIcon(QIcon(":/icons/privacy-mode"));
    }

    // Reload the website.
    webEngineViewPointer->reload();
}

void BrowserView::updateInterface() const
{
    // Update the URL line edit if it does not have focus.  TODO.  Remove block.
    if (!urlLineEditPointer->hasFocus())
    {
        // Update the URL line edit.
        urlLineEditPointer->setText(webEngineViewPointer->url().toString());
    }

    // Update the URL line edit.
    emit updateUrlLineEdit(webEngineViewPointer->url().toString());

    // Update the status of the forward and back buttons.
    backButtonPointer->setEnabled(webEngineHistoryPointer->canGoBack());  // TODO Remove.
    emit updateBackAction(webEngineHistoryPointer->canGoBack());
    forwardButtonPointer->setEnabled(webEngineHistoryPointer->canGoForward());  // TODO Remove.
    emit updateForwardAction(webEngineHistoryPointer->canGoForward());

    // Reapply the zoom factor.  This is a bug in QWebEngineView that resets the zoom with every load.  Hopefully it will be fixed in Qt6.  <https://bugreports.qt.io/browse/QTBUG-51992>
    webEngineViewPointer->setZoomFactor(Settings::zoomFactor());
}
